import logging

from pandas import DataFrame

from autogluon.common.features.feature_metadata import FeatureMetadata

from .abstract import AbstractFeatureGenerator

logger = logging.getLogger(__name__)


class IdentityFeatureGenerator(AbstractFeatureGenerator):
    """IdentityFeatureGenerator simply passes the data along without alterations."""

    def _fit_transform(self, X: DataFrame, **kwargs) -> (DataFrame, dict):
        X_out = self._transform(X)
        return X_out, self.feature_metadata_in.type_group_map_special

    def _transform(self, X: DataFrame) -> DataFrame:
        return X

    @staticmethod
    def get_default_infer_features_in_args() -> dict:
        return dict()

    def _more_tags(self):
        return {"feature_interactions": False}

    def estimate_output_feature_metadata(self, feature_metadata_in: FeatureMetadata, **kwargs) -> FeatureMetadata:
        features_to_remove = feature_metadata_in.get_features(**self._infer_features_in_args)
        return feature_metadata_in.keep_features(features_to_remove, inplace=False)



from typing import Dict, List, Literal, Tuple

import numpy as np
import pandas as pd
from sklearn.preprocessing import OrdinalEncoder

from autogluon.common.features.types import (
    R_CATEGORY,
    R_OBJECT,
)

from .abstract import AbstractFeatureGenerator


class CatAsNumFeatureGenerator(AbstractFeatureGenerator):
    """
    Convert object/category columns:
      - If a column's non-null values are all numeric-like, cast it with pd.to_numeric.
      - Otherwise, apply an OrdinalEncoder (per-column) to map categories to integers.

    Parameters
    ----------
    target_type : str
        The type of the target variable ('regression', 'binary', or 'multiclass').
    keep_original : bool, default=False
        Whether to keep the original columns alongside the converted ones.
    handle_unknown : {"use_encoded_value", "error"}, default="use_encoded_value"
        Passed to each internal OrdinalEncoder.
    unknown_value : int or float, default=-1
        Value to use for unknown categories when handle_unknown="use_encoded_value".
    dtype : numpy dtype, default=np.float64
        Output dtype for numeric-like conversions and encoded columns.
    """

    def __init__(
        self,
        target_type: Literal["regression", "binary", "multiclass"],
        keep_original=False,
        handle_unknown: str = "use_encoded_value",
        unknown_value: float | int = -1,
        dtype=np.float64,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.target_type = target_type
        self.keep_original = keep_original
        self.handle_unknown = handle_unknown
        self.unknown_value = unknown_value
        self.dtype = dtype

    def estimate_no_of_new_features(self, X: pd.DataFrame, **kwargs) -> int:
        # TODO: Add dtype of new features to estimation for all preprocessors
        if self.keep_original:
            return X.select_dtypes(include=["object", "category"]).shape[1]
        else:  # TODO: Implement proper handling
            return X.select_dtypes(include=["object", "category"]).shape[1]

    def _fit(self, X_in: pd.DataFrame, y_in=None):
        X = X_in.copy()
        self.input_columns_: List[str] = list(X.columns)

        obj_cat_cols = X.select_dtypes(include=[R_OBJECT, R_CATEGORY]).columns.tolist()
        self.numeric_like_cols_: List[str] = []
        self.ordinal_cols_: List[str] = []
        self.passthrough_cols_: List[str] = [col for col in X.columns if col not in obj_cat_cols]
        self.encoders_: Dict[str, OrdinalEncoder] = {}

        # Decide column-by-column and fit encoders where needed
        for col in obj_cat_cols:
            s = X[col]
            # Determine if all non-null values can be parsed as numbers
            non_null = s.dropna()
            num_convertible = pd.to_numeric(non_null, errors="coerce").notna().all()

            if num_convertible:
                self.numeric_like_cols_.append(col)
            else:
                self.ordinal_cols_.append(col)
                enc = OrdinalEncoder(
                    handle_unknown=self.handle_unknown,
                    unknown_value=self.unknown_value,
                    dtype=self.dtype,
                )
                # Fit on a 2D array
                enc.fit(s.astype("category").to_frame())
                self.encoders_[col] = enc

        return self

    def _fit_transform(self, X: pd.DataFrame, y: pd.Series, **kwargs) -> Tuple[pd.DataFrame, dict]:
        self._fit(X, y, **kwargs)
        X_out = self._transform(X)
        return X_out, dict()

    def _transform(self, X_in: pd.DataFrame, **kwargs) -> pd.DataFrame:
        X_out = pd.DataFrame(index=X_in.index)

        # 1) Cast numeric-like string/category columns
        for col in self.numeric_like_cols_:
            if col in X_in.columns:
                X_out[col] = pd.to_numeric(X_in[col], errors="coerce").astype(self.dtype)

        # 2) Ordinal-encode the remaining categorical columns (per-column encoders)
        for col in self.ordinal_cols_:
            if col in X_in.columns:
                enc = self.encoders_[col]
                # transform expects 2D
                X_out[col] = enc.transform(X_in[[col]]).astype(self.dtype).ravel()

        X_out.columns = [f"{col}_num" for col in X_out.columns]
        if self.keep_original:
            X_out = pd.concat([X_in, X_out], axis=1)
        else:
            X_out = pd.concat([X_in[self.passthrough_cols_], X_out], axis=1)
        return X_out

    @staticmethod
    def get_default_infer_features_in_args() -> dict:
        return dict()



import re

import numpy as np
import pandas as pd

"""
Further filtering ideas:
- based on target_corr - if its the same, the feature is likely to contain the same info
"""


def remove_mostlynan_features(X: pd.DataFrame, nan_threshold: float = 0.99) -> pd.DataFrame:
    return X.loc[:, X.isna().mean() < nan_threshold]


def remove_imbalanced(X: pd.DataFrame, mode_imbalance_threshold=0.99) -> pd.DataFrame:
    feature_imbalance = X.apply(mode_freq_fast)
    return X.loc[:, feature_imbalance < mode_imbalance_threshold]


def mode_freq_fast(s: pd.Series) -> float:
    vals, counts = np.unique(s.to_numpy(), return_counts=True)
    return counts.max() / len(s)


def remove_same_range_features(X: pd.DataFrame, x: pd.Series) -> float:
    # TODO: Change function name to be more descriptive of what actually happens
    col = x.name
    feature_names = [f for f in re.split(r"_(\*|/|\+|\-)_", col) if f not in {"*", "/", "+", "-"}]

    return X[feature_names].corrwith(x, method="spearman").max()


def basic_filter(
    X: pd.DataFrame,
    # y: pd.Series,
    min_cardinality: int = 3,
    candidate_cols: list = None,
    use_polars: bool = False,
    data_cleaning: bool = True,
    nan_threshold: float = 0.99,
    mode_imbalance_threshold: float = 0.99,
) -> pd.DataFrame:
    """
    Basic filtering of base and generated features:
    - Remove features with cardinality < min_cardinality
    - Optionally remove features that are constant or mostly NaN
    Parameters
    ----------
    X : pd.DataFrame
        Input features to filter
    min_cardinality : int, default=3
        Minimum cardinality required to keep a feature
    candidate_cols : list, default=None
        If provided, only these columns will be considered for filtering
    use_polars : bool, default=False
        Whether to use Polars for filtering (faster for large datasets)
    remove_constant_mostlynan : bool, default=True
        Whether to remove features that are nearly constant or mostly NaN
    Returns
    -------
    pd.DataFrame
        Filtered features
    """
    # Filter by minimum cardinality
    X = X.loc[:, X.nunique() >= min_cardinality]

    if X.empty:
        return X

    # if predetermined candidate columns are given, use them
    if candidate_cols is not None:
        X = X[candidate_cols]

    if data_cleaning:
        X = remove_mostlynan_features(X, nan_threshold=nan_threshold)
        X = remove_imbalanced(X, mode_imbalance_threshold=mode_imbalance_threshold)

    return X


def fast_spearman(X: pd.DataFrame) -> pd.DataFrame:
    """Compute Spearman correlation matrix using rank transformation and pairwise Pearson correlation."""
    # 1) Rank in pandas to match tie handling exactly
    R = X.rank(method="average", na_option="keep")
    A = R.to_numpy(float)
    # A = R.to_numpy(dtype=np.float32) # Could be float32 for less memory, but float64 is more accurate
    from ._numba_opt import _pearson_pairwise_nan

    C = _pearson_pairwise_nan(A)  # numba-accelerated pairwise corr
    return pd.DataFrame(C, index=X.columns, columns=X.columns)


def drop_high_corr(corr: pd.DataFrame, corr_threshold: float = 0.9) -> list:
    """Identify columns to drop based on a correlation threshold."""
    ac = corr.abs().copy()
    np.fill_diagonal(ac.values, 0.0)
    upper = ac.where(np.triu(np.ones(ac.shape), k=1).astype(bool))
    return upper.gt(corr_threshold).any(axis=0)[lambda s: s].index.tolist()


def filter_by_spearman(X: pd.DataFrame, corr_threshold: float = 0.95) -> list:
    """Filtering of a base feature set based on correlation."""
    spearman_corr = fast_spearman(X)
    np.fill_diagonal(spearman_corr.values, 0)
    drop_cols = drop_high_corr(spearman_corr, corr_threshold=corr_threshold)
    return [col for col in X.columns if col not in drop_cols]


def cross_spearman(df1, df2):
    """
    Compute Spearman correlations between all features in df1 vs df2.
    Missing values allowed. Pairwise-complete observations used.
    """
    # ---- 1) Rank-transform dataframes ----
    r1 = df1.rank(axis=0, na_option="keep")
    r2 = df2.rank(axis=0, na_option="keep")

    r1 = r1.to_numpy(float)
    r2 = r2.to_numpy(float)

    # ---- 2) Masks for non-missing ranks ----
    mask1 = ~np.isnan(r1)
    mask2 = ~np.isnan(r2)

    # ---- 3) Center ranks but keep NaN in place ----
    # Need nanmean to avoid biasing mean when NaNs present
    r1_centered = r1 - np.nanmean(r1, axis=0)
    r2_centered = r2 - np.nanmean(r2, axis=0)

    # Replace NaNs with zero so dot products ignore them
    r1c = np.where(mask1, r1_centered, 0.0)
    r2c = np.where(mask2, r2_centered, 0.0)

    # ---- 4) Compute pairwise sample sizes (valid rows per pair) ----
    # Broadcast masks: (samples × n1 × n2)
    n_valid = mask1[:, :, None] & mask2[:, None, :]
    n_valid = n_valid.sum(axis=0)  # shape (n1, n2)

    # ---- 5) Compute covariance using dot product but divide by valid count-1 ----
    cov = (r1c.T @ r2c) / (n_valid - 1)

    # ---- 6) Compute std dev for each feature using valid data ----
    # std = sqrt( sum((x-mean)^2) / (n_valid-1) )
    ss1 = (r1c**2).sum(axis=0)  # sum of squares for each column
    ss2 = (r2c**2).sum(axis=0)

    std1 = np.sqrt(ss1 / (mask1.sum(axis=0) - 1))  # shape (n1,)
    std2 = np.sqrt(ss2 / (mask2.sum(axis=0) - 1))

    # Outer-normalize to get Spearman correlations
    corr = cov / np.outer(std1, std2)

    # Where n_valid < 2 → correlation is undefined
    corr[n_valid < 2] = np.nan

    return pd.DataFrame(corr, index=df1.columns, columns=df2.columns)


def filter_by_cross_correlation(X_base: pd.DataFrame, X_new: pd.DataFrame, corr_threshold: float = 0.95) -> list:
    """Filter new features by cross-correlation with base features."""
    cross_corr = cross_spearman(X_base, X_new)
    to_drop = set()
    for new_col in X_new.columns:
        if (cross_corr[new_col].abs() > corr_threshold).any():
            to_drop.add(new_col)
    to_keep = [col for col in X_new.columns if col not in to_drop]
    novelty_scores = 1 - cross_corr[to_keep].abs().max()  # Higher --> more novel
    return to_keep, novelty_scores



from collections import defaultdict
from itertools import combinations
from math import comb
from typing import List, Literal, Tuple

import numpy as np
import pandas as pd

from autogluon.common.features.types import (
    R_CATEGORY,
    R_OBJECT,
    S_DATETIME_AS_OBJECT,
    S_IMAGE_BYTEARRAY,
    S_IMAGE_PATH,
)
from autogluon.features.generators.category import CategoryFeatureGenerator

from .abstract import AbstractFeatureGenerator
from .frequency import FrequencyFeatureGenerator


class CategoricalInteractionFeatureGenerator(AbstractFeatureGenerator):
    """
    Generate new categorical features by combining existing categorical features.
    Parameters
    ----------
    target_type : str
        The type of the target variable ('regression' or 'classification').
    max_order : int, default=2
        The maximum order of interactions to generate.
    max_new_feats : int, default=100
        Maximum number of new features to generate.
    candidate_cols : list of str or None, default=None
        List of candidate columns to consider for interaction generation. If None, all categorical columns are used.
    add_freq : bool, default=False
        Whether to add frequency encoding for the new features.
    only_freq : bool, default=False
        Whether to only keep frequency encoded features.
    min_cardinality : int, default=6
        Minimum cardinality of categorical features to be considered for interaction generation.
    min_count : int, default=2
        Minimum count of the most frequent category in a new feature to be retained.
    fillna : int, default=0
        Value to fill NaNs in frequency encoding.
    log : bool, default=False
        Whether to apply log transformation in frequency encoding.
    random_state : int, default=42
        Random state for reproducibility.
    **kwargs
        Additional keyword arguments.
    Returns
    -------
    self : CategoricalInteractionFeatureGenerator
        Fitted CategoricalInteractionFeatureGenerator instance.
    """

    def __init__(
        self,
        target_type: Literal["regression", "multiclass", "binary"],
        max_order: int = 3,
        max_new_feats: int = 100,
        candidate_cols: List[str] = None,
        add_freq: bool = False,
        only_freq: bool = False,
        min_cardinality: int = 2,
        min_count: int = 2,
        fillna: int = 0,
        log: bool = False,
        random_state: int = 42,
        inference_mode: Literal["string", "category"] = "category",
        make_categoricals: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.__name__ = "CatIntAdder"
        self.max_order = max_order
        self.candidate_cols = candidate_cols
        self.add_freq = add_freq
        self.only_freq = only_freq
        self.min_cardinality = min_cardinality
        self.min_count = min_count
        self.fillna = fillna
        self.log = log
        self.max_new_feats = max_new_feats
        self.inference_mode = inference_mode

        self.rng = np.random.default_rng(random_state)

        self.new_dtypes = {}

        self.make_categoricals = make_categoricals
        self.fitted_ = False
        self.cat_sizes_ = {}  # base categorical cardinalities
        self.reverse_mapping_ = {}  # interaction_col → sorted unique int combos

    # FIXME: Implement passthrough a bit better -> If feature in output that is also in input, drop the input feature?
    # TODO: This is the correct one, but due to the `passthrough` logic, need to keep all.
    #  Need to refactor passthrough logic to be more nuanced in order to enable this
    @staticmethod
    def get_default_infer_features_in_args() -> dict:
        return dict(
            valid_raw_types=[R_OBJECT, R_CATEGORY],
            invalid_special_types=[S_DATETIME_AS_OBJECT, S_IMAGE_PATH, S_IMAGE_BYTEARRAY],
            # required_raw_special_pairs=[
            #     (R_BOOL, None),
            #     (R_OBJECT, None),
            #     (R_CATEGORY, None),
            #     # (R_INT, S_BOOL),
            #     # (R_FLOAT, S_BOOL),
            # ],
        )

    # @staticmethod
    # def get_default_infer_features_in_args() -> dict:
    #     return dict()

    def estimate_new_dtypes(self, n_numeric, n_categorical, n_binary, **kwargs) -> int:
        num_new_feats = 0
        for order in range(2, self.max_order + 1):
            if n_categorical < order:
                continue
            num_new_feats += comb(n_categorical, order)

            if num_new_feats >= self.max_new_feats:
                num_new_feats = self.max_new_feats
                break

        if self.only_freq:
            return n_numeric + num_new_feats, n_categorical, n_binary
        elif self.add_freq:
            return n_numeric + num_new_feats, n_categorical + num_new_feats, n_binary
        else:
            return n_numeric, n_categorical + num_new_feats, n_binary

    def estimate_no_of_new_features(self, X: pd.DataFrame, **kwargs) -> int:
        """
        Estimate the number of new features that will be generated.
        Parameters
        ----------
        X : pd.DataFrame
            Input DataFrame.
        Returns
        -------
        int
            Estimated number of new features.
        Notes
        -----
        This is a rough estimate based on the number of categorical features
        that pass the cardinality filter and the specified maximum order.
        Currently, doesn't consider the min_count filter. Therefore, the actual
        number of generated features may be lower than this estimate.
        """
        X_cat = X.select_dtypes(include=["object", "category"])
        pass_cardinality_filter = X_cat.nunique().values >= self.min_cardinality
        num_base_feats = np.sum(pass_cardinality_filter)
        affected_features = X_cat.columns[pass_cardinality_filter].tolist()

        num_new_feats = 0
        for order in range(2, self.max_order + 1):
            if num_base_feats < order:
                continue
            num_new_feats += comb(num_base_feats, order)

        return np.min([self.max_new_feats, num_new_feats]), affected_features

    def group_feat_combs_by_order(self, new_col_set):
        feat_combs = [s.split("_&_") for s in new_col_set]
        by_order = defaultdict(list)
        for cols in feat_combs:
            by_order[len(cols)].append(cols)
        return dict(by_order)

    def get_interaction_names_dict(self, X: pd.DataFrame, max_order: int = 2, max_feats: int = 100):
        self.new_col_set = []
        for o in range(2, max_order + 1):
            feat_combs_use = list(combinations(np.unique(X.columns), o))
            add_cols = ["_&_".join(f) for f in feat_combs_use]
            max_feats_order = max_feats - len(self.new_col_set)
            if len(add_cols) > max_feats_order:
                add_cols = self.rng.choice(add_cols, max_feats_order, replace=False).tolist()
            self.new_col_set.extend(add_cols)
            if len(self.new_col_set) >= max_feats:
                self.new_col_set = self.new_col_set[:max_feats]
                break
        self.new_cols_by_order = self.group_feat_combs_by_order(self.new_col_set)

        self.used_cols = []
        for col in self.new_col_set:
            self.used_cols.extend(col.split("_&_"))
        self.used_cols = list(set(self.used_cols))

    def combine_predefined(self, X_in: pd.DataFrame, comb_lst: List[str], fit_mode=True, **kwargs) -> pd.DataFrame:
        """Generate interaction features based on predefined combinations.
        Parameters
        ----------
        X_in : pd.DataFrame
            Input categorical features DataFrame.
        comb_lst : List[str]
            List of predefined combinations as strings. Can be higher order combinations, but one call to this function should only handle one order at a time.
        Returns
        -------
        pd.DataFrame
            DataFrame containing the generated interaction features.
        """
        X = X_in.copy()
        X = X_in.astype("U")
        feat_combs_use = [i.split("_&_") for i in comb_lst]
        feat_combs_use_arr = np.array(feat_combs_use).transpose()

        features = X[feat_combs_use_arr[0]].values
        for num, arr in enumerate(feat_combs_use_arr[1:]):
            features += "_&_" + X[arr].values

        X_out = pd.DataFrame(features, columns=comb_lst, index=X.index)

        return X_out

    # ----------------------------------------------------------------------
    # INTERNAL: build interactions (faster version)
    # ----------------------------------------------------------------------
    def _build_interactions(self, X, fit_mode):
        # Precompute base codes (faster than repeatedly calling .cat.codes)
        base_codes = {col: X[col].cat.codes.to_numpy(np.int32) for col in X.columns}

        outputs = {}  # column_name → numpy array

        # --------------------------
        # ORDER 2 interactions
        # --------------------------
        if 2 in self.new_cols_by_order:
            for cols in self.new_cols_by_order[2]:
                name = "_&_".join(cols)
                col1, col2 = cols
                k = self.cat_sizes_[col2]

                combined = base_codes[col1] * k + base_codes[col2]
                outputs[name] = self._fit_or_map(name, combined, fit_mode)

                # Expose newly created interaction codes for higher-order use
                base_codes[name] = outputs[name]

        # --------------------------
        # ORDER >= 3 interactions
        # --------------------------
        for order in sorted(k for k in self.new_cols_by_order if k >= 3):
            for cols in self.new_cols_by_order[order]:
                name = "_&_".join(cols)

                prefix = "_&_".join(cols[:-1])
                last = cols[-1]

                k = self.cat_sizes_[last]
                combined = base_codes[prefix] * k + base_codes[last]

                outputs[name] = self._fit_or_map(name, combined, fit_mode)
                base_codes[name] = outputs[name]

        # ------------------------------------------------------------
        # Build final DataFrame in one fast pass
        # ------------------------------------------------------------
        if not outputs:
            return pd.DataFrame(index=X.index)

        df_out = pd.DataFrame(outputs, index=X.index)

        # Convert to categoricals with consistent ordering
        if self.make_categoricals:
            for col in df_out.columns:
                uniques = self.reverse_mapping_[col]
                df_out[col] = pd.Categorical.from_codes(
                    df_out[col],
                    categories=range(len(uniques)),  # ensures 0..n-1
                    ordered=False,
                )

        return df_out

    # ----------------------------------------------------------------------
    # INTERNAL: fit or transform mapping (FAST version)
    # ----------------------------------------------------------------------
    def _fit_or_map(self, name, combined, fit_mode):
        """
        combined: int64 array of encoded interaction IDs
        """

        if fit_mode:
            valid = combined[combined >= 0]
            uniques = np.unique(valid)
            self.reverse_mapping_[name] = uniques
            return self._vectorized_map(combined, uniques)
        else:
            uniques = self.reverse_mapping_[name]
            return self._vectorized_map(combined, uniques)

    # ----------------------------------------------------------------------
    # INTERNAL: vectorized mapping using np.searchsorted (FAST)
    # ----------------------------------------------------------------------
    @staticmethod
    def _vectorized_map(combined, uniques):
        """
        Map combined IDs to 0..(n-1) using vectorized searchsorted.
        Unseen combos → -1.
        """
        mapped = np.full(combined.shape, -1, dtype=np.int64)

        valid_mask = combined >= 0
        vals = combined[valid_mask]

        # vectorized position lookup
        idx = np.searchsorted(uniques, vals)

        # Step 1: out-of-range positions (unseen)
        oob = idx == len(uniques)

        # Step 2: in-range but not equal → unseen mismatch
        in_range = ~oob
        mismatch = np.zeros_like(oob)
        mismatch[in_range] = uniques[idx[in_range]] != vals[in_range]

        miss = oob | mismatch

        # Assign -1 to misses, otherwise idx
        out = np.full(vals.shape, -1, dtype=np.int64)
        out[~miss] = idx[~miss]

        mapped[valid_mask] = out
        return mapped.astype(np.int32)

    def frequency_encode_new_features(self, X: pd.DataFrame, y: pd.Series = None):
        if self.add_freq or self.only_freq:
            if not self.fitted_:
                # TODO: Unclear whether there is a more efficient way to do this
                candidate_cols = FrequencyFeatureGenerator.filter_candidates_by_distinctiveness(X[self.new_col_set])
                if len(candidate_cols) > 0:
                    keep_original = not self.only_freq
                    self.cat_freq = FrequencyFeatureGenerator(
                        candidate_cols=candidate_cols, fillna=self.fillna, log=self.log, keep_original=keep_original
                    )
                    return self.cat_freq.fit_transform(X[candidate_cols], y)
                else:
                    return X
            else:
                return self.cat_freq.transform(X)
        else:
            return X

    def _fit(self, X: pd.DataFrame, y: pd.Series, **kwargs):
        if self.candidate_cols is None:
            self.candidate_cols = X.select_dtypes(include=["object", "category"]).columns.tolist()
            self.candidate_cols = [
                i for i in self.candidate_cols if X[i].nunique() >= self.min_cardinality
            ]  # TODO: Make this a parameter

        # FIXME: why `self.max_order` and not 2?
        if len(self.candidate_cols) < self.max_order:
            self.new_col_set = []
            return self

        X = X[self.candidate_cols].copy()
        # Ensure categorical only when needed
        for col in self.candidate_cols:
            if not isinstance(X[col].dtype, pd.CategoricalDtype):
                X[col] = X[col].astype("category")

        self.get_interaction_names_dict(X[self.candidate_cols], max_order=self.max_order, max_feats=self.max_new_feats)

        # Cache category sizes (+1 for missing index -1)
        self.cat_sizes_ = {col: len(X[col].cat.categories) + 1 for col in self.used_cols}

        self.fitted_ = True

        return self

    # FIXME: Filter if too unique?
    def _fit_transform(self, X: pd.DataFrame, y: pd.Series, **kwargs) -> Tuple[pd.DataFrame, dict]:
        self._fit(X, y, **kwargs)
        X_out = self._transform(X, fit_mode=True)

        # features_out = list(X_out.columns)

        # type_group_map_special = {R_CATEGORY: features_out} # TODO: Find out whether that is needed
        return X_out, dict()

    def _transform(self, X: pd.DataFrame, y: pd.Series = None, fit_mode: bool = False, **kwargs) -> pd.DataFrame:
        if len(self.new_col_set) == 0:
            return pd.DataFrame(index=X.index)

        if self.inference_mode == "string":
            X_out = pd.DataFrame(index=X.index)
            for degree in range(2, self.max_order + 1):
                col_set_use = [col for col in self.new_col_set if col.count("_&_") + 1 == degree]
                if len(col_set_use) > 0:
                    X_degree = self.combine_predefined(X[self.used_cols], col_set_use, fit_mode=fit_mode)
                    X_out = pd.concat([X_out, X_degree], axis=1)
            if fit_mode:
                self.cat_transformer = CategoryFeatureGenerator(minimum_cat_count=1)
                X_out = self.cat_transformer.fit_transform(X_out)
            else:
                X_out = self.cat_transformer.transform(X_out)

        else:  # new inference mode
            for col in self.used_cols:
                if not isinstance(X[col].dtype, pd.CategoricalDtype):
                    X[col] = X[col].astype("category")
            X_out = self._build_interactions(X[self.used_cols], fit_mode=fit_mode)

        X_out = self.frequency_encode_new_features(X_out)

        return X_out



"""
Code adapted from skrub==0.6.2
"""

from __future__ import annotations

import sklearn
from sklearn.utils.fixes import parse_version

sklearn_version = parse_version(parse_version(sklearn.__version__).base_version)


if sklearn_version < parse_version("1.6"):

    def validate_data(_estimator, /, **kwargs):
        if "ensure_all_finite" in kwargs:
            force_all_finite = kwargs.pop("ensure_all_finite")
        else:
            force_all_finite = True
        return _estimator._validate_data(**kwargs, force_all_finite=force_all_finite)
else:
    from sklearn.utils.validation import validate_data  # noqa: F401
